/*
 * cluster.h - Primary header file for cluster program(s).
 *
 * Copyright (C) 1994 Baylor College of Medicine
 *
 * All rights reserved worldwide.
 *
 * Contact Brent Wiese, brent@bcm.tmc.edu
 */

/*****************
 * This is the header file for the cluster program.  The following program
 * has rather complex data structures and as such they will be briefly
 * described here.
 *
 * This program was derived from the cluster program written by
 * Donald V. Faulkner, may he rest in peace.
 *
 * Please note that the "n" used below refers to the maximum number of nodes
 * for the given clustering run.
 *
 * Name		Range		Note
 * ----		-----		----
 * cluster	[1,n]		Refers to a group of nodes that are
 *				similar, which is what clustering is.
 * node		n		A individual item/object that is being
 *				clustered.  Technically this should be
 *				called a vertex according to graph theory.
 * edge		[0,n*n/4-1]	A relationship between two nodes with a
 *				corresponding weight/score.  There are
 *				actually (n*n-n)/2 possible edges but since
 *				edges of merged clusters/nodes are deleted,
 *				the maximum is less than the actually value
 *				and is only obtainable using maximal linkage.
 *				For single linkage the value is always zero
 *				since no edges need to be remembered.
 * link		[n,2*n-1]	A link used to relate nodes in a cluster
 *				that includes the weight/score.
 *
 *
 * Some additional things should be mentioned.  The clustering uses a
 * minimum score value as a threshold but the actual score values are
 * never compared except to throw away values.  As such it is imperative
 * that the edges be sorted accordingly prior to clustering with this
 * routine.
 */

#ifndef clusterDefined
#define clusterDefined

#include <config.h>
#include <nodeSet.h>
#include <edgeSet.h>
#include <scoreType.h>

/* The printEdges(1) program removed the need for include and exclude files
 * in this program.
 */
/*#define IncludeExcludeFiles*/	/* Use a exclude and include file. */
/*#define IncludeAnd*/		/* Require both nodes of edge to be in list. */
/*#define OutputFullLink*/		/* Printout full link else just score. */

typedef struct clusterExtern	clusterExternType;
typedef struct cluster	clusterType;
typedef struct node	nodeType;
typedef struct edge	edgeType;
typedef struct link	linkType; /* This is a cluster link rather than a
				   * node link (edge) */
typedef int	mergeTestFuncType(clusterType* aCluster,
				  clusterType* bCluster,
				  edgeType* edge);
typedef void	deleteEdgesFuncType(clusterType* aCluster,
				    clusterType* bCluster);
typedef void	linkageFuncType(void);
typedef int	(*qsortFuncType)(const void *a, const void *b);

/*****************
 * Generate node set with: "cSet -c counter i node r nodeId i none p"
 *   -or-		   "cSet -c counter i node r nodeId s none p"
 * Generate edge set with: "cSet -c counter i edge r none p none v"
 * Generate score type with: "cType score i"
 *   -or-		     "cType score f"
 *
 * where the counter is a common generic counter value, nodeId is either
 * numeric or string, and the value of a the node set is a pointer to a node
 * which is also the key of a edge set.  Additionally the score type is one
 * of integer or float.
 */

struct cluster {
	nodeType	*nodes;	/* Linked list of nodes. */
	nodeType	*lastNode; /* Last node in linked list. */
	counterType	n;	/* Number of nodes in linked list. */
	linkType	*root;	/* Root of linked cluster tree. */
	clusterType	*next;	/* Next cluster in doubly linked list. */
	clusterType	*prev;	/* Previous cluster in doubly linked list. */
};

struct node {
	nodeIdType	id;	/* Node identifier. */
	clusterType	*memberOf; /* Cluster containing this node. */
	nodeType	*next;	/* Next node in linked list. */
	edgeSetType	edges;	/* Set of edges from (but not to) this node. */
};

struct edge {
	nodeIdType	a;	/* First/left node of edge. */
	nodeIdType	b;	/* Second/right node of edge. */
	scoreType	score;	/* Weight/score of edge join the two nodes. */
};

struct link {
	linkType	*left;	/* Left cluster link. */
	linkType	*right;	/* Right cluster link. */
	scoreType	score;	/* Weight/score when clusters were linked. */
#ifdef OutputFullLink
	nodeIdType	a;	/* First/left node. */
	nodeIdType	b;	/* Second/right node. */
#endif /* OutputFullLink */
};

/*****************
 * Make sure that any changes made here are made to the initialization
 * of clusterExtern below.  THEY MUST REMAIN IN SYNC.  Ideally a initialization
 * routine should be called, but at the time automatic initialization seems
 * most appropriate.
 */
struct clusterExtern {
	clusterType	nilBuf;	/* Buffer for nil (head/tail) of circular
				   doubly linked list.  Ideally suited for
				   insertion and deletion since no checks
				   for empty/beginning/end of list are
				   required. */
	clusterType	*nil;	/* Circular doubly Linked list of clusters
				   including a nil cluster. */
	counterType	n;	/* Number of clusters. */
	counterType	nNodes;	/* Number of nodes in all clusters. */
	FILE		*edgeFile; /* FILE to read edges from. */
	FILE		*clusterFile; /* FILE to write cluster to. */
	counterType	minimumNodes; /* Minimum number of nodes to print. */
	int		dumpOrder; /* <0 reverse sort, >0 sort, 0 no sort. */
	scoreType	minimumScore; /* Minimum score to accept link. */
	mergeTestFuncType*	mergeTest; /* Test for need to merge. */
	deleteEdgesFuncType*	deleteEdges; /* Routine to delete edges. */
	int		linkageProportion; /* Percent of links needed for
					    * proportional linkage. */
	counterType	linkageAbsolute; /* Number of absolute links needed for
					  * absolute type linkage. */
};

#if defined(clusterVariables) && !defined(clusterVariablesDefined)
#define clusterVariablesDefined

clusterExternType	clusterExternBuf = {
	{
		(nodeType *) 0,	/* nodes */
		(nodeType *) 0, /* lastNode */
		(counterType) 0, /* n */
		(linkType *) 0,	/* root */
		&clusterExternBuf.nilBuf, /* next (first) */
		&clusterExternBuf.nilBuf, /* prev (last) */
	},			/* nilBuf */
	&clusterExternBuf.nilBuf, /* nil */
	(counterType) 0,	/* n */
	(counterType) 0,	/* nNodes */
	stdin,			/* edgeFile */
	stdout,			/* clusterFile */
	(counterType) 2,	/* minimumNodes */
	(int) -1,		/* dumpOrder */
	(scoreType) 5,		/* minimumScore */
	0,			/* mergeTest */
	0,			/* deleteEdges */
	50,			/* linkageProportion */
	3,			/* linkageAbsolute */
	};
clusterExternType	*clusterExtern = &clusterExternBuf;

#else /* clusterVariables */

extern clusterExternType	*clusterExtern;

#endif /* clusterVariables */

int       clusterCmp(clusterType *a, clusterType *b);
void      doDump(void);
int       dumpCmp(clusterType **a, clusterType **b);
int       dumpCmpReverse(clusterType **a, clusterType **b);
void      dumpTree(linkType *link);
void		merge(clusterType *a, clusterType *b, edgeType *edge);
nodeType	*newNode(void);
void		singleLinkage(void);
void		linkage(void);
mergeTestFuncType		mergeTestComplete;
mergeTestFuncType		mergeTestProportional;
mergeTestFuncType		mergeTestAbsoluteOrComplete;
deleteEdgesFuncType		deleteEdgesGeneric;
int		nextEdge(edgeType* edge);
void		nextEdgeInit(char* includeFileName, char* excludeFileName);
int		readEntrezBin(edgeType *edge);
void    	stop(char *s);

#endif /* clusterDefined */
