/*
 * mergeTest.c - Test if last edge brought in will merge two clusters.
 *
 * Copyright (C) 1994 Baylor College of Medicine
 *
 * All rights reserved worldwide.
 *
 * Contact Brent Wiese, brent@bcm.tmc.edu
 */

#include <cluster.h>

/*****************
 * Merge the two clusters if all the nodes (vertices) have links (edges).
 */

int	mergeTestComplete(clusterType* aCluster,
			  clusterType* bCluster,
			  edgeType* edge)
{
	nodeType*	aNode;
	nodeType*	bNode;
	edgeSetType*	edgeSet;
	nodeType*	key;

	aNode = aCluster->nodes;
	while (aNode != 0)
	{
		bNode = bCluster->nodes;
		while (bNode != 0)
		{
			if (
#ifdef nodeCmp
			    nodeCmp(aNode->id, bNode->id) > 0
#else
			    aNode->id > bNode->id
#endif
			    )
			{
				edgeSet = &bNode->edges;
				key = aNode;
			}
			else
			{
				edgeSet = &aNode->edges;
				key = bNode;
			}

			if (edgeSearch(edgeSet, (void *) key) == 0)
			{
				return(0);
			}
			bNode = bNode->next;
		}
		aNode = aNode->next;
	}
	return(1);
}

/*****************
 * Merge the two clusters if a certain portion (percent) of the nodes (sum of
 * the number of nodes in A and B) have edges (links).
 */

int	mergeTestProportional(clusterType* aCluster,
			      clusterType* bCluster,
			      edgeType* edge)
{
	nodeType*	aNode;
	nodeType*	bNode;
	edgeSetType*	edgeSet;
	nodeType*	key;
	counterType	matches; /* Number of edges matched so far. */
	counterType	possibleMatches; /* Number of tests left + matches. */
	counterType	neededMatches; /* Number of matches needed for this
					* test to be true. */

	matches = 0;
	possibleMatches = aCluster->n * bCluster->n;
	neededMatches = (possibleMatches * clusterExtern->linkageProportion
			 + 50) / 100;

	aNode = aCluster->nodes;
	while (aNode != 0)
	{
		bNode = bCluster->nodes;
		while (bNode != 0)
		{
			if (
#ifdef nodeCmp
			    nodeCmp(aNode->id, bNode->id) > 0
#else
			    aNode->id > bNode->id
#endif
			    )
			{
				edgeSet = &bNode->edges;
				key = aNode;
			}
			else
			{
				edgeSet = &aNode->edges;
				key = bNode;
			}

			if (edgeSearch(edgeSet, (void *) key) != 0)
			{
				matches++;
				if (matches >= neededMatches)
				{
					return(1);
				}
			}
			else
			{
				possibleMatches--;
				if (possibleMatches < neededMatches)
				{
					return(0);
				}
			}
			bNode = bNode->next;
		}
		aNode = aNode->next;
	}

	/* We should never get here. */
	stop("cluster: ran through mergeTestProportional() ????\n");
	return(0);
}

/*****************
 * Merge the two clusters if there are atleast an absolute number of links
 * (edges), or the clusters are completely linked.  The last qualification
 * is necessary unless the matrix of edges needs to be read in prior to
 * doing any clustering, otherwise no clusters would ever be merged.
 * The true absolute linkage would be similar to a UPGMA implementation which
 * has not been addressed at this time.
 */

int	mergeTestAbsoluteOrComplete(clusterType* aCluster,
				    clusterType* bCluster,
				    edgeType* edge)
{
	nodeType*	aNode;
	nodeType*	bNode;
	edgeSetType*	edgeSet;
	nodeType*	key;
	counterType	matches; /* Number of edges matched so far. */
	counterType	possibleMatches; /* Number of tests left + matches. */
	counterType	neededMatches; /* Number of matches needed for this
					* test to be true. */

	matches = 0;
	possibleMatches = aCluster->n * bCluster->n;
	neededMatches = clusterExtern->linkageAbsolute;
	if (neededMatches > possibleMatches)
	{
		neededMatches = possibleMatches;
	}

	aNode = aCluster->nodes;
	while (aNode != 0)
	{
		bNode = bCluster->nodes;
		while (bNode != 0)
		{
			if (
#ifdef nodeCmp
			    nodeCmp(aNode->id, bNode->id) > 0
#else
			    aNode->id > bNode->id
#endif
			    )
			{
				edgeSet = &bNode->edges;
				key = aNode;
			}
			else
			{
				edgeSet = &aNode->edges;
				key = bNode;
			}

			if (edgeSearch(edgeSet, (void *) key) != 0)
			{
				matches++;
				if (matches >= neededMatches)
				{
					return(1);
				}
			}
			bNode = bNode->next;
		}
		aNode = aNode->next;
	}
	return(0);
}
