/*
 * new_read.c - Function that create/initialize a sequence file for reading.
 *
 * Copyright (C) 1994 Baylor College of Medicine
 *
 * All rights reserved.
 *
 * Contact Brent Wiese, brent@bcm.tmc.edu
 */

#include <config.h>
#include <memu_macros.h>
#include <sequ.h>
#include <sequ_lib.h>

/*
 * This function reads the first sequence from a multiple sequence file.
 * It automatically senses the format of the sequence and will return
 * a pointer to the sequence file information, in addition to assigning
 * to a pointer passed as an argument the sequence sequence information.
 *
 * If fp is NULL, then be sure to open filename.  Return NULL at EOF.
 */

sequ_file_t *
sequ_new_read_first(sequ_seq_t **seq_p, char *filename, FILE *fp)
{
    sequ_file_t	*file;
    sequ_seq_t	*seq;
    char	*line;
    size_t	line_length;


    memu_new_m(file, sequ_file_t, 1);

    if (filename == NULL)
	file->filename = "noname";
    else
	file->filename = filename;

    /* Make sure we can return the sequence sequence also. */
    if (seq_p == NULL) {
	fprintf(stderr,
		"%s: unexpected null sequence pointer for file %s\n",
		program_name,
		file->filename);
	abort();
    }

    /* open filename? */
    if (fp == NULL) {
	fp = fopen(file->filename, "r");
	if (fp == NULL) {
	    fprintf(stderr,
		    "%s: error opening file %s\n",
		    program_name,
		    file->filename);
	    perror(program_name);
	    abort();
	}
    }
    file->fp = fp;
    file->format = sequ_format_unknown;
    file->last_line = NULL;
    file->last_line_length = 0;
    file->did_first = FALSE;

    /* read a line and determine the format */
    line = _sequ_getline(fp, &line_length);
    if (line == NULL) {
	/* one might consider this an error condition since the sequence
	 * format can not be determined from nothing, but since many
	 * unix utilities do nothing with no input, just return the file
	 * as is but assign to the sequence pointer a null sequence.
	 */
	*seq_p = NULL;
	return(file);
    }

    file->last_line = line;
    file->last_line_length = line_length;

    /* Let's allocate a sequence (seq) and initialize so this can be
     * avoided in all the actual format specific read functions.
     */
    memu_new_m(seq, sequ_seq_t, 1);
    seq->id = NULL;
    seq->seq = NULL;
    seq->length = 0;
    seq->title = NULL;

    switch(line[0]) {
    case '>':
	if (line[3] == ';') {
	    file->format = sequ_format_nbrf;
	    _sequ_read_nbrf(seq, file);
	} else {
	    file->format = sequ_format_fasta;
	    _sequ_read_fasta(seq, file);
	}
	break;
    case ';':
	/* Currently this only works if a ";" style comment line is present. */
	file->format = sequ_format_ig;
	_sequ_read_ig(seq, file);
	break;
    case 'L':
	if (strncmp(line, "LOCUS ", 6) != CMP_EQUAL)
	    break;
	file->format = sequ_format_gb;
	_sequ_read_gb(seq, file);
	break;
    case 'I':
	if (strncmp(line, "ID ", 3) != CMP_EQUAL)
	    break;
	file->format = sequ_format_embl;
	_sequ_read_embl(seq, file);
	break;
    case '\\':
	if (strcmp(line, "\\\\\\") != CMP_EQUAL)
	    break;
	file->format = sequ_format_pir;
	/* Get the next line which is the start of the first sequence. */
	memu_free_m(file->last_line);
	file->last_line = _sequ_getline(fp, &file->last_line_length);
	if (file->last_line == NULL) {
	    *seq_p = NULL;
	    return(file);
	}
	file->did_first = TRUE;
	_sequ_read_pir(seq, file);
	break;

	/* No default */
    }

    /* None single character auto sensing. */
    if (file->format == sequ_format_unknown) {

	/* The tbl format is the only one at this time.  Since the tbl
	 * format is really only defined by a tab character, just make
	 * sure there is atleast one tab in the line.  This library is not
	 * intended to check the integrity of your sequence file so this
	 * should be sufficient.
	 */

	char *tab;

	tab = strchr(line, '\t');
	if (tab != NULL) {
	    file->format = sequ_format_tbl;
	    _sequ_read_tbl(seq, file);
	}
    }

    if (file->format == sequ_format_unknown) {
	fprintf(stderr,
		"%s: unable to determine sequence file format\n",
		program_name);
	abort();
    }

    /* This indicated is set to true for all the formats in which
     * it makes no difference but want it set for good measure.
     */
    file->did_first = TRUE;
    *seq_p = seq;
    return(file);
}
