/*
 * read.c - Functions that read sequence files.
 *
 * Copyright (C) 1994 Baylor College of Medicine
 *
 * All rights reserved.
 *
 * Contact Brent Wiese, brent@bcm.tmc.edu
 */

#include <config.h>
#include <ctype.h>
#include <memu_macros.h>
#include <sequ.h>
#include <sequ_lib.h>


/* Really nasty macros, only use a single statement with no side effects
 * with the arguments.
 */
#define skip_spaces_m(p) while (*(p) && isspace(*(p)) != FALSE) p++
#define skip_till_space_m(p) while (*(p) && isspace(*(p)) == FALSE) p++

static char *bad_line_follows = "offending line follows\n";

/*
 * Read the next sequence from a file of known format.
 */

sequ_seq_t *
sequ_read(sequ_file_t *file)
{
    sequ_seq_t	*seq;

    /* Read the next line if we haven't already. */
    if (file->last_line == NULL)
	file->last_line = _sequ_getline(file->fp, &file->last_line_length);

    /* Might be end-of-file */
    if (file->last_line == NULL)
	return(NULL);

    /* Let's allocate a sequence (seq) and initialize so this can be
     * avoided in every format specific read function.
     */
    memu_new_m(seq, sequ_seq_t, 1);
    seq->id = NULL;
    seq->seq = NULL;
    seq->length = 0;
    seq->title = NULL;

    switch(file->format) {
    case sequ_format_tbl:
	_sequ_read_tbl(seq, file);
	break;
    case sequ_format_fasta:
	_sequ_read_fasta(seq, file);
	break;
    case sequ_format_ig:
	_sequ_read_ig(seq, file);
	break;
    case sequ_format_gb:
	_sequ_read_gb(seq, file);
	break;
    case sequ_format_nbrf:
	_sequ_read_nbrf(seq, file);
	break;
    case sequ_format_embl:
	_sequ_read_embl(seq, file);
	break;
    case sequ_format_pir:
	_sequ_read_pir(seq, file);
	break;
    default:
	fprintf(stderr,
		"%s: must specify known format for file %s\n",
		program_name,
		file->filename);
	abort();
	break;
    }

    return(seq);
}


/*
 * Read tbl format.  Just split the line into two or three tab separated
 * fields, assigning the extra field to title if present.  The id is the
 * first field and the sequence the last, the title is anything between.
 */

void
_sequ_read_tbl(sequ_seq_t *seq, sequ_file_t *file)
{
    char *first_tab;
    char *last_tab;
    size_t m;
    size_t n;

    first_tab = strchr(file->last_line, '\t');
    if (first_tab == NULL) {
	fprintf(stderr,
		"%s: Missing tab in tbl sequence from %s, %s",
		program_name,
		file->filename,
		bad_line_follows);
	fprintf(stderr, "%s: %s\n", program_name, file->last_line);
	abort();
    }

    m = first_tab - file->last_line;
    seq->id = _sequ_ndup(file->last_line, m);

    /* We know there is always a last tab since we know there was a first
     * tab and the last tab can be the same as the first tab.
     */
    last_tab = strrchr(file->last_line, '\t');

    n = last_tab - first_tab;
    if (n > 0) {
	/* Don't count the possible last tab that isn't the same as the
	 * first tab.
	 */
	seq->title = _sequ_ndup(first_tab + 1, n - 1);
    }

    /* Do not count the first tab in the length. */
    m = file->last_line_length - m - 1 - n; /* == strlen(last_tab + 1) */
    seq->seq = _sequ_ndup(last_tab + 1, m);
    seq->length = m;

    memu_free_m(file->last_line);
    file->last_line = NULL;
}


/*
 * Read fasta format.  First line starts with '>' with id immediately
 * following it until first space, title follows that.  Sequence lines
 * until line starting with '>'.
 */

void
_sequ_read_fasta(sequ_seq_t *seq, sequ_file_t *file)
{
    char *p;
    char *q;
    size_t m;
    size_t length;
    char *line;
    size_t line_length;

    if (file->last_line[0] != '>') {
	fprintf(stderr,
		"%s: Missing '>' in fasta sequence on %d, %s",
		program_name,
		fileno(file->fp),
		bad_line_follows);
	fprintf(stderr, "%s: %s\n", program_name, file->last_line);
	abort();
    }

    /* Include the '>' character in the length, count it as the null. */
    p = q = file->last_line + 1;
    skip_till_space_m(p);
    seq->id = _sequ_ndup(q, p - q);

    skip_spaces_m(p);
    if (*p != '\0') {
	m = file->last_line_length - (p - file->last_line);
	seq->title = _sequ_ndup(p, m);
    }

    memu_free_m(file->last_line);

    length = 0;
    p = NULL;
    do {
	line = _sequ_getline(file->fp, &line_length);
	if (line == NULL || line[0] == '>')
	    break;
	if (p == NULL) {
	    p = line;
	    length = line_length;
	} else {
	    if (line_length != 0) {
		/* A little picky, but why jump through loops to copy
		 * a zero length string.
		 */
		memu_resize_m(p, char, length + line_length + 1);
		strcpy(p + length, line);
		length += line_length;
	    }
	    /* getline() will allocate a byte for the null even if line is
	     * a blank line (no characters).
	     */
	    memu_free_m(line);
	}
    } while(TRUE);

    file->last_line = line;
    file->last_line_length = line_length;

    if (p == NULL) {
	seq->seq = _sequ_ndup("", 0);
    } else {
	seq->seq = p;
    }
    seq->length = length;
}


/*
 * Read IG format.  Check first line comment (that might contain possible
 * title but going to ignore it for now), skip comment lines that start ';',
 * id next line by itself on the line, sequences lines till comment line,
 * deleting possible '1' or '2' at the end of the sequence.
 */

void
_sequ_read_ig(sequ_seq_t *seq, sequ_file_t *file)
{
    char *p;
    size_t m;
    size_t length;
    char *line;
    size_t line_length;

    if (file->last_line[0] != ';') {
	fprintf(stderr,
		"%s: Missing ';' in ig sequence on %d, %s",
		program_name,
		fileno(file->fp),
		bad_line_follows);
	fprintf(stderr, "%s: %s\n", program_name, file->last_line);
	abort();
    }

    p = file->last_line + 1;
    skip_spaces_m(p);
    if (p != '\0') {
	m = file->last_line_length - (p - file->last_line);
	seq->title = _sequ_ndup(p, m);
    }
    memu_free_m(file->last_line);
    file->last_line = _sequ_getline(file->fp, &file->last_line_length);

    while (file->last_line[0] == ';') {
	memu_free_m(file->last_line);
	file->last_line = _sequ_getline(file->fp, &file->last_line_length);

	/* The following should never really happen but ... */
	if (file->last_line == NULL) {
	    fprintf(stderr,
		    "%s: IG format file has comments without sequence\n",
		    program_name);
	    abort();
	}
    }

    seq->id = file->last_line;	/* Do not free this one. */

    length = 0;
    p = NULL;
    do {
	line = _sequ_getline(file->fp, &line_length);
	if (line == NULL || line[0] == ';')
	    break;
	if (p == NULL) {
	    p = line;
	    length = line_length;
	} else {
	    if (line_length != 0) {
		/* A little picky, but why jump through loops to copy
		 * a zero length string.
		 */
		memu_resize_m(p, char, length + line_length + 1);
		strcpy(p + length, line);
		length += line_length;
	    }
	    /* getline() will allocate a byte for the null even if line is
	     * a blank line (no characters).
	     */
	    memu_free_m(line);
	}
    } while(TRUE);

    file->last_line = line;
    file->last_line_length = line_length;

    if (length > 0 &&
	(p[length - 1] == '1' || p[length - 1] == '2')) {

	p[length - 1] = '\0';	/* This wastes one byte but not going to
					 * worry about it. */
	    length--;
    }

    if (p == NULL) {
	seq->seq = _sequ_ndup("", 0);
    } else {
	seq->seq = p;
    }
    seq->length = length;
}


/* Read GenBank format.  Check first line contain "LOCUS" and stash id right
 * after that.  Skip lines through "ORIGIN" line, stashing possible "TITLE"
 * line, and sequence follows the "ORIGIN" line until "//" line.  Must
 * remove initial numbering from sequence lines and blank every 10 positions.
 */

void
_sequ_read_gb(sequ_seq_t *seq, sequ_file_t *file)
{
    char *p;
    char *q;
    char *r;
    size_t m;
    size_t length_in_file;
    size_t length;
    char *line;
    size_t line_length;
    int num_full_groups;

    /****
    if (strncmp(file->last_line, "LOCUS       ", 12) != CMP_EQUAL) {
     ****/
    if (strncmp(file->last_line, "LOCUS ", 6) != CMP_EQUAL) {
	fprintf(stderr,
	       "%s: Missing 'LOCUS' line in GB sequence on %d, %s",
	       program_name,
	       fileno(file->fp),
	       bad_line_follows);
	fprintf(stderr, "%s: %s\n", program_name, file->last_line);
	abort();
    }

    p = file->last_line + 6;
    skip_spaces_m(p);
    q = p;
    skip_till_space_m(p);
    seq->id = _sequ_ndup(q, p - q);

    skip_spaces_m(p);
    if (isdigit(*p))
	length_in_file = atol(p);
    else
	length_in_file = 0;

    memu_free_m(file->last_line);

    do {
	line = _sequ_getline(file->fp, &line_length);
	if (line == NULL || strncmp(line, "ORIGIN", 6) == CMP_EQUAL)
	    break;
	if (strncmp(line, "DEFINITION ", 11) == CMP_EQUAL) {
	    p = line + 11;
	    skip_spaces_m(p);
	    seq->title = _sequ_ndup(p, line_length - (p - line));
	}
	memu_free_m(line);
    } while(TRUE);

    if (line == NULL) {
	fprintf(stderr,
		"%s: GB format file has no sequence\n",
		program_name);
	abort();
    }
    memu_free_m(line);

    length = 0;
    p = NULL;
    do {
	line = _sequ_getline(file->fp, &line_length);
	if (line == NULL || strcmp(line, "//") == CMP_EQUAL)
	    break;

	/* Figure out how long the actual sequence data is. */
	if (line_length < 10)
	    abort();
	m = line_length - 10;
	num_full_groups = m / 11;
	m -= num_full_groups;

	if (p == NULL) {
	    memu_new_m(p, char, m + 1);
	} else {
	    memu_resize_m(p, char, length + m + 1);
	}

	q = line + 10;
	r = p + length;
	length += m;

	while(*q) {
	    if (isspace(*q) == FALSE) {
		*r = *q;
		r++;
	    }
	    q++;
	}
	*r = '\0';
	memu_free_m(line);
    } while(TRUE);

    if (line == NULL)
	abort();

    memu_free_m(line);

    file->last_line = NULL;

    if (length_in_file != 0 && length_in_file != length) {
	fprintf(stderr,
		"%s: inconsistent length for GB sequence\n",
		program_name);
	abort();
    }

    if (p == NULL) {
	seq->seq = _sequ_ndup("", 0);
    } else {
	seq->seq = p;
    }
    seq->length = length;
}


/*
 * Read NBRF format.  Check first line has '>' in first position and ';'
 * in the fourth position, with the ID right after that.  The next line
 * is a comment with sequence lines following that until blank line.
 * Deleting possible '*' at the end of the sequence.
 */

void
_sequ_read_nbrf(sequ_seq_t *seq, sequ_file_t *file)
{
    /* This reading function is different in that it assumes spaces
     * may or may not exist in the sequence data.
     */
    char *p;
    char *q;
    char *r;
    char *s;
    size_t m;
    size_t length;
    char *line;
    size_t line_length;

    if (file->last_line[0] != '>'
	|| file->last_line_length < 4
	|| file->last_line[3] != ';') {

	fprintf(stderr,
		"%s: Missing '>__;ID' in NBRF sequence on %d, %s",
		program_name,
		fileno(file->fp),
		bad_line_follows);
	fprintf(stderr, "%s: %s\n", program_name, file->last_line);
	abort();
    }

    p = q = file->last_line + 4;
    skip_till_space_m(p);
    seq->id = _sequ_ndup(q, p - q);
    memu_free_m(file->last_line);

    /* Title line. */
    line = _sequ_getline(file->fp, &line_length);
    if (line == NULL)
	abort();
    p = line;
    skip_spaces_m(p);
    if (*p != '\0') {
	seq->title = _sequ_ndup(p, line_length - (p - line));
    }
    memu_free_m(line);

    length = 0;
    p = NULL;
    do {
	line = _sequ_getline(file->fp, &line_length);
	if (line == NULL || line[0] == '\0' || line[0] == '>')
	    break;

	memu_new_m(s, char, line_length);
	q = line;
	r = s;
	while(*q) {
	    if (isspace(*q) == FALSE) {
		*r = *q;
		r++;
	    }
	    q++;
	}
	*r = '\0';
	m = r - s;

	if (p == NULL) {
	    memu_new_m(p, char, m + 1);
	} else {
	    memu_resize_m(p, char, length + m + 1);
	}

	strcpy(p + length, s);
	length += m;
	memu_free_m(line);
	memu_free_m(s);
    } while(TRUE);

    /* Why is this here, shouldn't a null last line be kept? */
    if (line != NULL && line[0] == '\0') {
	file->last_line = NULL;
    } else {
	file->last_line = line;
	file->last_line_length = line_length;
    }

    if (length > 0 && p[length - 1] == '*') {
	p[length - 1] = '\0';	/* This wastes one byte but not going to
				 * worry about it. */
	length--;
    }

    if (p == NULL) {
	seq->seq = _sequ_ndup("", 0);
    } else {
	seq->seq = p;
    }
    seq->length = length;
}


/* Read EMBL format.  Check first line contains "ID" and stash id right
 * after that.  Skip lines until "SQ" line, stashing length of sequence
 * at the end of the line.  Sequence lines until "//" line, blank every
 * 10 positions.
 */

void
_sequ_read_embl(sequ_seq_t *seq, sequ_file_t *file)
{
    char *p;
    char *q;
    char *r;
    size_t m;
    size_t length_in_file;
    size_t length;
    char *line;
    size_t line_length;
    int num_full_groups;

    if (strncmp(file->last_line, "ID ", 3) != CMP_EQUAL) {
	fprintf(stderr,
	       "%s: Missing 'ID' line in EMBL sequence on %d, %s",
	       program_name,
	       fileno(file->fp),
	       bad_line_follows);
	fprintf(stderr, "%s: %s\n", program_name, file->last_line);
	abort();
    }

    p = file->last_line + 3;
    skip_spaces_m(p);
    q = p;
    skip_till_space_m(p);
    seq->id = _sequ_ndup(q, p - q);
    memu_free_m(file->last_line);

    do {
	line = _sequ_getline(file->fp, &line_length);
	if (line == NULL || strncmp(line, "SQ ", 3) == CMP_EQUAL)
	    break;
	if (strncmp(line, "DE ", 3) == CMP_EQUAL) {
	    p = line + 3;
	    skip_spaces_m(p);
	    seq->title = _sequ_ndup(p, line_length - (p - line));
	}
	memu_free_m(line);
    } while(TRUE);

    if (line == NULL) {
	fprintf(stderr,
		"%s: EMBL format file has no sequence\n",
		program_name);
	abort();
    }

    p = line + 3;
    skip_spaces_m(p);
    if (isdigit(*p))
	length_in_file = atol(p);
    else
	length_in_file = 0;

    memu_free_m(line);

    length = 0;
    p = NULL;
    do {
	line = _sequ_getline(file->fp, &line_length);
	if (line == NULL || strcmp(line, "//") == CMP_EQUAL)
	    break;

	/* Figure out how long the actual sequence data is. */
	if (line_length < 5)
	    abort();
	m = line_length - 5;
	num_full_groups = m / 11;
	m -= num_full_groups;

	if (p == NULL) {
	    memu_new_m(p, char, m + 1);
	} else {
	    memu_resize_m(p, char, length + m + 1);
	}

	q = line + 5;
	r = p + length;
	length += m;

	while(*q) {
	    if (isspace(*q) == FALSE) {
		*r = *q;
		r++;
	    }
	    q++;
	}
	*r = '\0';
	memu_free_m(line);
    } while(TRUE);

    if (line == NULL)
	abort();

    memu_free_m(line);

    file->last_line = NULL;

    if (length_in_file != 0 && length_in_file != length) {
	fprintf(stderr,
		"%s: inconsistent length for EMBL sequence\n",
		program_name);
	abort();
    }

    if (p == NULL) {
	seq->seq = _sequ_ndup("", 0);
    } else {
	seq->seq = p;
    }
    seq->length = length;
}


/* Read PIR format.  Check first line is "\\\" for the first sequence and
 * "///" for all but the first, next line contains "ENTRY"
 * and stash id right after that.  Next line is possible "TITLE" line and
 * stash title right after that.  Skip lines through "SEQUENCE" line.  Skip
 * next line that has numbers only.  Sequence lines until "///" line,
 * blank every other position.
 */

void
_sequ_read_pir(sequ_seq_t *seq, sequ_file_t *file)
{
    char *p;
    char *q;
    char *r;
    size_t m;
    size_t length;
    char *line;
    size_t line_length;


    if (file->did_first == FALSE) {
	if (strcmp(file->last_line, "\\\\\\") != CMP_EQUAL) {
	    fprintf(stderr,
		    "%s: Missing initial '\\\\\\' line in PIR sequence on %d, %s",
		    program_name,
		    fileno(file->fp),
		    bad_line_follows);
	    fprintf(stderr, "%s: %s\n", program_name, file->last_line);
	    abort();
	}
	memu_free_m(file->last_line);
	file->last_line = _sequ_getline(file->fp, &file->last_line_length);
	/* This should never happen, but just in case. */
	if (file->last_line == NULL)
	    abort();
    }

    if (strncmp(file->last_line, "ENTRY ", 6) != CMP_EQUAL) {
	fprintf(stderr,
	       "%s: Missing 'ENTRY' line in PIR sequence on %d, %s",
	       program_name,
	       fileno(file->fp),
	       bad_line_follows);
	fprintf(stderr, "%s: %s\n", program_name, file->last_line);
	abort();
    }

    p = file->last_line + 6;
    skip_spaces_m(p);
    q = p;
    skip_till_space_m(p);
    seq->id = _sequ_ndup(q, p - q);
    memu_free_m(file->last_line);

    do {
	line = _sequ_getline(file->fp, &line_length);
	if (line == NULL || strncmp(line, "SEQUENCE", 8) == CMP_EQUAL)
	    break;
	if (strncmp(line, "TITLE ", 6) == CMP_EQUAL) {
	    p = line + 6;
	    skip_spaces_m(p);
	    seq->title = _sequ_ndup(p, line_length - (p - line));
	}
	memu_free_m(line);
    } while(TRUE);

    if (line == NULL) {
	fprintf(stderr,
		"%s: PIR format file has no sequence\n",
		program_name);
	abort();
    }

    memu_free_m(line);

    /* Skip blank line, should probably check that is was blank. */
    line = _sequ_getline(file->fp, NULL);
    if (line == NULL)
	abort();

    length = 0;
    p = NULL;
    do {
	line = _sequ_getline(file->fp, &line_length);
	if (line == NULL || strcmp(line, "///") == CMP_EQUAL)
	    break;

	/* Figure out how long the actual sequence data is. */
	if (line_length < 8)
	    abort();
	m = (line_length - 8) / 2;

	if (p == NULL) {
	    memu_new_m(p, char, m + 1);
	} else {
	    memu_resize_m(p, char, length + m + 1);
	}

	q = line + 8;
	r = p + length;
	length += m;

	while(*q) {
	    q++;
	    *r = *q;
	    r++;
	    q++;
	}
	*r = '\0';
	memu_free_m(line);
    } while(TRUE);

    if (line == NULL)
	abort();

    memu_free_m(line);

    file->last_line = NULL;

    if (p == NULL) {
	seq->seq = _sequ_ndup("", 0);
    } else {
	seq->seq = p;
    }
    seq->length = length;
}
