/*
 * write.c - Functions that write sequence files.
 *
 * Copyright (C) 1994 Baylor College of Medicine
 *
 * All rights reserved.
 *
 * Contact Brent Wiese, brent@bcm.tmc.edu
 */

#include <config.h>
#include <sequ.h>
#include <sequ_lib.h>


/*
 * Write the next sequence from a file of known format.
 */

void
sequ_write(sequ_file_t *file, sequ_seq_t *seq)
{
    if (seq == NULL || file == NULL)
	abort();

    switch(file->format) {
    case sequ_format_tbl:
	_sequ_write_tbl(file, seq);
	break;
    case sequ_format_fasta:
	_sequ_write_fasta(file, seq);
	break;
    case sequ_format_ig:
	_sequ_write_ig(file, seq);
	break;
    case sequ_format_gb:
	_sequ_write_gb(file, seq);
	break;
    case sequ_format_nbrf:
	_sequ_write_nbrf(file, seq);
	break;
    case sequ_format_embl:
	_sequ_write_embl(file, seq);
	break;
    case sequ_format_pir:
	_sequ_write_pir(file, seq);
	break;
    default:
	fprintf(stderr,
		"%s: must specify known format for file %s\n",
		program_name,
		file->filename);
	abort();
	break;
    }

    if (ferror(file->fp) != FALSE) {
	fprintf(stderr,
		"%s: write error on %d\n",
		program_name,
		fileno(file->fp));
	abort();
    }
    file->did_first = TRUE;
}


/*
 * All of the functions below are the inverse of the read function.
 * To understand what these functions are writing out it would be
 * best to look at the comments before each read function that is
 * the inverse of the write function here.
 */

void
_sequ_write_tbl(sequ_file_t *file, sequ_seq_t *seq)
{
    fputs(seq->id, file->fp);
    putc('\t', file->fp);
    if (seq->title != NULL) {
	fputs(seq->title, file->fp);
	putc('\t', file->fp);
    }
    fputs(seq->seq, file->fp);
    putc('\n', file->fp);
}


void
_sequ_write_fasta(sequ_file_t *file, sequ_seq_t *seq)
{
    char *p;
    int column;

    putc('>', file->fp);
    fputs(seq->id, file->fp);
    if (seq->title != NULL) {
	putc(' ', file->fp);
	fputs(seq->title, file->fp);
    }
    putc('\n', file->fp);

    for (p = seq->seq, column = 1; *p != '\0'; p++, column++) {
	if (column > FASTA_WIDTH) {
	    putc('\n', file->fp);
	    column = 1;
	}
	putc(*p, file->fp);
    }
    putc('\n', file->fp);
}


void
_sequ_write_ig(sequ_file_t *file, sequ_seq_t *seq)
{
    char *p;
    int column;

    putc(';', file->fp);
    if (seq->title != NULL) {
#if 0 /* Might consider putting a space here but the format does not say so. */
	putc(' ', file->fp);
#endif
	fputs(seq->title, file->fp);
    }
    putc('\n', file->fp);
    fputs(seq->id, file->fp);
    putc('\n', file->fp);

    for (p = seq->seq, column = 1; *p != '\0'; p++, column++) {
	if (column > IG_WIDTH) {
	    putc('\n', file->fp);
	    column = 1;
	}
	putc(*p, file->fp);
    }

    /* The following doesn't seem to be the convention, so for the last
     * line the line will exceed the usual width if the length of the
     * sequence is a multiple of the width (plus spaces).
     */
#if 0
    if (column > IG_WIDTH) {
	putc('\n', file->fp);
    }
#endif
    putc('1', file->fp);
    putc('\n', file->fp);
}


void
_sequ_write_gb(sequ_file_t *file, sequ_seq_t *seq)
{
    size_t n;
    char *p;
    int column;
    int group;

    fputs("LOCUS       ", file->fp);
    fputs(seq->id, file->fp);
    fputs("       ", file->fp);
    fprintf(file->fp, "%ld bp\n", (long) seq->length);
    if (seq->title != NULL) {
	fputs("DEFINITION  ", file->fp);
	fputs(seq->title, file->fp);
	putc('\n', file->fp);
    }
    fputs("ORIGIN      \n", file->fp);

    n = 1;
    fprintf(file->fp, "%8ld  ", (long) n);
    for (p = seq->seq, column = 1, group = 1;
	 *p != '\0';
	 p++, column++, group++, n++) {
	if (column > GB_WIDTH) {
	    fprintf(file->fp, "\n%8ld  ", (long) n);
	    column = 1;
	    group = 1;
	}
	if (group > GB_GROUP) {
	    putc(' ', file->fp);
	    group = 1;
	}
	putc(*p, file->fp);
    }
    fputs("\n//\n", file->fp);
}


void
_sequ_write_nbrf(sequ_file_t *file, sequ_seq_t *seq)
{
    char *p;
    int column;
    int group;

    /* Should really check that a sequence is P1 at some point. */
    fputs(">P1;", file->fp);
    fputs(seq->id, file->fp);
    putc('\n', file->fp);
    if (seq->title) {
	fputs(seq->title, file->fp);
    }
    putc('\n', file->fp);

#if 0
    /* Careful, printing trailing space that starts next line. */
    fprintf(file->fp, ", %ld bases\n ", (long) seq->length);
#else
    putc(' ', file->fp);
#endif

    for (p = seq->seq, column = 1, group = 1;
	 *p != '\0';
	 p++, column++, group++) {
	if (column > NBRF_WIDTH) {
	    putc('\n', file->fp);
	    putc(' ', file->fp);
	    column = 1;
	    group = 1;
	}
	if (group > NBRF_GROUP) {
	    putc(' ', file->fp);
	    group = 1;
	}
	putc(*p, file->fp);
    }

    /* The following doesn't seem to be the convention, so for the last
     * line the line will exceed the usual width if the length of the
     * sequence is a multiple of the width (plus spaces).
     */
#if 0
    if (column > NBRF_WIDTH) {
	putc('\n', file->fp);
	putc(' ', file->fp);
	group = 1;
    }
    if (group > NBRF_GROUP) {
	putc(' ', file->fp);
    }
#endif
    putc('*', file->fp);
    putc('\n', file->fp);
    putc('\n', file->fp);
}


void
_sequ_write_embl(sequ_file_t *file, sequ_seq_t *seq)
{
    char *p;
    int column;
    int group;

    fputs("ID   ", file->fp);
    fputs(seq->id, file->fp);
    if (seq->title != NULL) {
	fputs("\nDE   ", file->fp);
	fputs(seq->title, file->fp);
    }
    fprintf(file->fp, "\nSQ             %ld BP\n", (long) seq->length);

    fputs("     ", file->fp);
    for (p = seq->seq, column = 1, group = 1;
	 *p != '\0';
	 p++, column++, group++) {
	if (column > EMBL_WIDTH) {
	    fputs("\n     ", file->fp);
	    column = 1;
	    group = 1;
	}
	if (group > EMBL_GROUP) {
	    putc(' ', file->fp);
	    group = 1;
	}
	putc(*p, file->fp);
    }
    fputs("\n//\n", file->fp);
}

void
_sequ_write_pir(sequ_file_t *file, sequ_seq_t *seq)
{
    size_t n;
    char *p;
    int column;

    if (file->did_first == FALSE)
	fputs("\\\\\\\n", file->fp);
    fputs("ENTRY           ", file->fp);
    fputs(seq->id, file->fp);
    /* Don't know if really supposed to put a space at the end of this line
     * but the output from readseq has it so I am going to put it there
     * for now.
     */
    putc(' ', file->fp);
    if (seq->title != NULL) {
	fputs("\nTITLE           ", file->fp);
	fputs(seq->title, file->fp);
    }
    fputs("\nSEQUENCE        \n", file->fp);
    fputs(PIR_HEADING, file->fp);

    n = 1;
    fprintf(file->fp, "%7ld ", (long) n);
    for (p = seq->seq, column = 1; *p != '\0'; p++, column++, n++) {
	if (column > PIR_WIDTH) {
	    fprintf(file->fp, "\n%7ld ", (long) n);
	    column = 1;
	}
	putc(' ', file->fp);
	putc(*p, file->fp);
    }
    fputs("\n///\n", file->fp);
}
