#include <pima.h>

#ifndef PIMA_HOME
#define PIMA_HOME "/usr/local/lib/pima"
#endif

#define COMMON_OPTIONS "ad:hi:l:m:w:"

#ifdef PIMA_MSO
#define GET_RESPONSES get_responses_mso
#define ALLOWED_OPTIONS COMMON_OPTIONS "x"
#define ARGUMENT_COUNT 3
#define PIMA_INDEX PIMA_MSO_INDEX
#endif

#ifdef PIMA_PM
#define ALLOWED_OPTIONS COMMON_OPTIONS "e:st:u:"
#define GET_RESPONSES get_responses_pm
#define ARGUMENT_COUNT 4
#define PIMA_INDEX PIMA_PM_INDEX
#endif

#ifndef GET_RESPONSES
#error One of PIMA_MSO, PIMA_PM, or PIMA_PLS must be defined.
#endif

extern char *optarg;
extern int optind;

static void print_help(void);

void
GET_RESPONSES(int argc, char **argv, pima_params_t *params)
{
    char input[MAX_FILENAME_LEN];
    char buffer[MAX_FILENAME_LEN];
    char datadir[MAX_FILENAME_LEN];
    int not_found;
    char *p;
    int c;
    char *matrix_file;
    FILE *fp;

    float m_coeff1, m_coeff2, m_coeff3, m_min_local_score;
    float o_coeff1, o_coeff2;
#ifdef PIMA_PM
    float o_coeff3;
#endif
    float o_min_local_score;
    float *coeff1_p, *coeff2_p, *coeff3_p, *min_local_score_p;

    int ask_for_options, ask_for_arguments, options_given;
    

    /* Use the environment variable PIMA_HOME as the datadir if set. */
    p = getenv("PIMA_HOME");
    if (p != NULL) {
	strcpy(datadir, p);
    } else {
	strcpy(datadir, PIMA_HOME);
    }
    strcat(datadir, "/");
	
    /* Initialize parameters with defaults. */
    params->main_index = PIMA_INDEX;
    params->sec_struct = FALSE;
    params->lex = FALSE;
    params->W2_atten = 1.0;
    params->min_local_width = 15;
    params->name_suffix = "";
    params->ss_chars = "hge";
    matrix_file = "patgen.mat";

    /* Setup pointers to values from matrix file so that if option is
     * specified we can just change the pointers, but still read the values
     * from the matrix file without concern.
     */

    ask_for_options = FALSE;
    ask_for_arguments = FALSE;
    options_given = FALSE;
    min_local_score_p = &m_min_local_score;
    coeff1_p = &m_coeff1;
    coeff2_p = &m_coeff2;
    coeff3_p = &m_coeff3;

    while ((c = getopt(argc, argv, ALLOWED_OPTIONS)) != EOF) {
	switch (c) {
	case 'a':
	    ask_for_options = TRUE;
	    break;
	case 'd':
	    options_given = TRUE;
	    o_coeff2 = atof(optarg);
	    coeff2_p = &o_coeff2;
	    break;
	case 'h':
	    print_help();
	    break;
	case 'i':
	    options_given = TRUE;
	    o_coeff1 = atof(optarg);
	    coeff1_p = &o_coeff1;
	    break;
	case 'l':
	    options_given = TRUE;
	    o_min_local_score = atof(optarg);
	    min_local_score_p = &o_min_local_score;
	    break;
	case 'm':
	    options_given = TRUE;
	    matrix_file = optarg;
	    break;
	case 'w':
	    options_given = TRUE;
	    params->min_local_width = atoi(optarg);
	    break;
#ifdef PIMA_MSO
	case 'x':
	    params->lex = TRUE;
	    break;
#endif
#ifdef PIMA_PM
	case 'e':
	    /* Use the option 'e' as in extension, but must include the dot. */
	    options_given = TRUE;
	    params->name_suffix = optarg;
	    break;
	case 's':
	    params->sec_struct = TRUE;
	    break;
	case 't':
	    options_given = TRUE;
	    o_coeff3 = atof(optarg);
	    coeff3_p = &o_coeff3;
	    break;
	case 'u':
	    options_given = TRUE;
	    params->ss_chars = optarg;
	    break;
#endif
	case '?':
	default:
	    fprintf(stderr, "%s: bad option '%c', run '%s -h' for help.\n",
		    program_name, c, program_name);
	    exit(1);
	    break;
	}
    }

    /* See if there are any arguments, that the argument count is good, and
     * then process the arguments.
     */
    if (optind >= argc) {
	/* Ask for options if there were no arguments nor any
	 * interesting options.
	 */
	if (options_given == FALSE) {
	    ask_for_options = TRUE;
	}
	ask_for_arguments = TRUE;

    } else if (optind + ARGUMENT_COUNT != argc) {
	fprintf(stderr, "%s: bad arguments, run '%s -h' for help.\n",
		program_name, program_name);
	exit(1);
    }	

    if (ask_for_options == TRUE) {
	/* Be sure to ask for and read the matrix file first so that the
	 * default values for the given matrix file can be presented.
	 */
	not_found = 0;
	do {
	    printf("Enter weight-matrix filename [%s]: ", matrix_file);
	    get_string_or_default(matrix_file, input);

	    /* If the matrix name doesn't contain any slashes, then assume
	     * they want a file in datadir.  If they want a file from the
	     * current directory then they can always type "./name" instead of
	     * just "name".
	     */
	    p = strchr(input, '/');
	    if (p == NULL) {
		strcpy(buffer, datadir);
		strcat(buffer, input);
		p = buffer;
	    } else {
		p = input;
	    }
	    fp = fopen(p, "r");
	    if (fp == NULL) {
		printf("Weight-matrix file \"%s\" not found, please try again.\n", p);
		not_found = 1;
	    }
	} while (not_found);
	read_matrix(fp, params->AA_chars, params->C, params->S, &params->n_AA,
		    &m_min_local_score, &m_coeff1, &m_coeff2);
	m_coeff3 = m_coeff1 * 10.0;

	printf("Enter minimum local score to continue alignment [%1.2f]: ",
	       *min_local_score_p);
	*min_local_score_p = get_float_or_default(*min_local_score_p);

	sprintf(buffer, "%d", params->min_local_width);
	printf("Enter minimum local width to continue alignment [%s]: ",
	       buffer);
	get_string_or_default(buffer, input);
	params->min_local_width = atoi(input);
	
	printf("Enter length independent gap penalty (gap open) [%1.2f]: ",
	       *coeff1_p);
	*coeff1_p = get_float_or_default(*coeff1_p);

	printf("Enter length dependent gap penalty (gap extension) [%1.2f]: ",
	       *coeff2_p);
	*coeff2_p = get_float_or_default(*coeff2_p);

#ifdef PIMA_PM
	if (params->sec_struct == TRUE) {
	    printf("Enter secondary structure dependent gap penalty [%1.2f]: ",
		   *coeff3_p);
	    *coeff3_p = get_float_or_default(*coeff3_p);

	    printf("Enter secondary structure characters to be gap constrained [%s]: ", params->ss_chars);
	    get_string_or_default(params->ss_chars, input);
	    params->ss_chars = strdup(input);
	    if (params->ss_chars == NULL) {
		fprintf(stderr, "%s: strdup of ss_chars.\n", program_name);
		abort();
	    }
	}

	printf("Enter locus name suffix");
	if (params->name_suffix[0] == '\0') {
	    printf(": ");
	} else {
	    printf(" [%s]: ", params->name_suffix);
	}
	get_string_or_default(params->name_suffix, input);
	params->name_suffix = strdup(input);
	if (params->name_suffix == NULL) {
	    fprintf(stderr, "%s: strdup of name_suffix.\n", program_name);
	    abort();
	}
#endif
    } else {			/* ask_for_options is FALSE */
	/* Possibly use datadir, see comment above. */
	p = strchr(matrix_file, '/');
	if (p == NULL) {
	    strcpy(buffer, datadir);
	    strcat(buffer, matrix_file);
	    p = buffer;
	} else {
	    p = matrix_file;
	}
	fp = fopen(p, "r");
	if (fp == NULL) {
	    fprintf(stderr, "%s: Open matrix file \"%s\".\n", program_name, p);
	    exit(1);
	}
	read_matrix(fp, params->AA_chars, params->C, params->S, &params->n_AA,
		    &m_min_local_score, &m_coeff1, &m_coeff2);
	m_coeff3 = m_coeff1 * 10.0;

    } /* End if ask_for_options is TRUE */

    params->min_local_score = (int) (*min_local_score_p * 100.0);
    params->coeff1 = (int) (*coeff1_p * 100.0);
    params->coeff2 = (int) (*coeff2_p * 100.0);
    params->coeff3 = (int) (*coeff3_p * 100.0);

    if (ask_for_arguments == TRUE) {
	do {
	    not_found = 0;
	    printf("Enter name of first input file: ");
	    get_string_or_default("", input);
	    fp = fopen(input, "r");
	    if (fp == NULL) {
		printf("Input file \"%s\" not found, please try again.\n",
		       input);
		not_found = 1;
	    }
	} while (not_found);
	params->pattern_filename = strdup(input);
	if (params->pattern_filename == NULL) {
	    fprintf(stderr, "%s: strdup of pattern_filename.\n", program_name);
	    abort();
	}
	params->pattern_file = fp;

	do {
	    not_found = 0;
	    printf("Enter name of second input file: ");
	    get_string_or_default("", input);
	    fp = fopen(input, "r");
	    if (fp == NULL) {
		printf("Input file \"%s\" not found, please try again.\n",
		       input);
		not_found = 1;
	    }
	} while (not_found);
	params->query_filename = strdup(input);
	if (params->query_filename == NULL) {
	    fprintf(stderr, "%s: strdup of query_filename.\n", program_name);
	    abort();
	}
	params->query_file = fp;

	do {
	    not_found = 0;
	    printf("Enter name of output file: ");
	    get_string_or_default("", input);
	    fp = fopen(input, "w");
	    if (fp == NULL) {
		printf("Unable to open output file \"%s\", please try again.\n", input);
		not_found = 1;
	    }
	} while (not_found);
	params->root_filename = strdup(input);
	if (params->root_filename == NULL) {
	    fprintf(stderr, "%s: strdup of root_filename.\n", program_name);
	    abort();
	}
	params->root_file = fp;

#ifdef PIMA_PM
	printf("Enter pattern name: ");
	get_string_or_default("", input);
	params->pattern_name = strdup(input);
	if (params->pattern_name == NULL) {
	    fprintf(stderr, "%s: strdup of pattern_name.\n", program_name);
	    abort();
	}
#endif
    } else {			/* if ask_for_parameters is FALSE */
	/* We already know the number of arguments is correct. */
	fp = fopen(argv[optind], "r");
	if (fp == NULL) {
	    fprintf(stderr, "%s: Open first input file \"%s\".\n",
		    program_name, argv[optind]);
	    exit(1);
	}
	params->pattern_filename = argv[optind];
	params->pattern_file = fp;
	optind++;

	fp = fopen(argv[optind], "r");
	if (fp == NULL) {
	    fprintf(stderr, "%s: Open second input file \"%s\".\n",
		    program_name, argv[optind]);
	    exit(1);
	}
	params->query_filename = argv[optind];
	params->query_file = fp;
	optind++;

	fp = fopen(argv[optind], "w");
	if (fp == NULL) {
	    fprintf(stderr, "%s: Open output file \"%s\".\n",
		    program_name, argv[optind]);
	    exit(1);
	}
	params->root_filename = argv[optind];
	params->root_file = fp;
	optind++;

#ifdef PIMA_PM
	params->pattern_name = argv[optind];
#endif
    }
}

static void
print_help(void)
{
#ifdef PIMA_MSO
    printf("Usage: pima-mso [options] first_input_file second_input_file output_file\n\n");
#endif
#ifdef PIMA_PM
    printf("Usage: pima-pm [options] first_input_file second_input_file output_file pattern_name\n\n");
#endif
    printf("  -a              ask for options even if options or arguments\n");
    printf("  -d PENALTY      use length dependent gap penalty of PENALTY (gap extension)\n");
#ifdef PIMA_PM
    printf("  -e SUFFIX       add suffix (extension) to sequence/pattern names.\n");
#endif
    printf("  -h              print this help\n");
    printf("  -i PENALTY      use length independent gap penalty of PENALTY (gap open)\n");
    printf("  -l SCORE        use minimum local score of SCORE\n");
    printf("  -m MATRIX_FILE  use weight matrix file of MATRIX_FILE instead of pima2.mat\n");
#ifdef PIMA_PM
    printf("  -s              assume data contains secondary structure information.\n");
    printf("  -t PENALTY      use sec. struct. gap penalty of PENALTY\n");
    printf("  -u CHARS        use secondary structure characters CHARS instead of \"hge\"\n");
#endif
    printf("  -w WIDTH        use minimum local alignment width of WIDTH instead of 15\n");
#ifdef PIMA_MSO
    printf("  -x              only calculate scores when in lexographical order\n");
#endif
    printf("\nDefault values for the gap penalties and the minimum local score depend on\n");
    printf("the weight matrix file used.");
#ifdef PIMA_PM
    printf("  The secondary structure gap penalty default\n");
    printf("value is 10 times the length independent gap penalty for the given weight matrix.");
#endif
    printf("\n");
    exit(0);
}
